/***************************************************************************
	mmpnm - Utilities for loading PNM (Portable Anymap) files

			-------------------
        begin           : 6/4/2004
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "mmpnm.h"

/*******************************************************************/

int mmpnm_framing::set_resolution(int new_resolution)
{
	if (new_resolution < 600)
	{
		fprintf(stderr, "Invalid resolution %ddpi: Resolution must be >= 600dpi\n", new_resolution);
		new_resolution = 600;
	}

	memset(this, 0, sizeof(*this)); 
	resolution_dpi = new_resolution;
	film_width = FILM_WIDTH_1200_DPI * resolution_dpi / 1200;
	frame_width = FRAME_WIDTH_1200_DPI * resolution_dpi / 1200;
	frame_height = FRAME_HEIGHT_1200_DPI * resolution_dpi / 1200;
	sprocket_height = SPROCKET_HEIGHT_1200_DPI * resolution_dpi / 1200;
	sprocket_width = SPROCKET_WIDTH_1200_DPI * resolution_dpi / 1200;
	sprocket_to_edge = SPROCKET_TO_EDGE_1200_DPI * resolution_dpi / 1200;
	sprocket_area_width = SPROCKET_AREA_WIDTH_1200_DPI * resolution_dpi / 1200;
	return 0;	
}


/*******************************************************************/

mmpnm_file::~mmpnm_file()
{
	if (color_data)
		delete color_data;
	if (gray_data)
		delete gray_data;
}

int mmpnm_file::add_flicker(int flicker_level)
{
	if ((flicker_level <= 1) || !this || !gray_data || !color_data)
		return -EINVAL;

	int level = random() % flicker_level;
	printf("Adding flicker: level %d\n", level);

	for (int y = 0; y < height; ++y)
		for (int x = 0; x < width; ++x)
			for (int color = -1; color < 3; ++color)
				add_pixel(level, x, y, color);

	return 0;
}

int mmpnm_file::add_grain(int grain_level, int grain_count)
{
	if ((grain_level < 1) || !this || !gray_data || !color_data)
		return -EINVAL;

	if (grain_count <= 0)
		grain_count = width * height * 2;
	
	printf("Adding grain: level %d, count %d\n", grain_level, grain_count);

	for (int count = 0; count < grain_count; ++count)
	{
		float x = (random() % (width * 256)) / 256.0;
		float y = (random() % (height * 256)) / 256.0;
		int level = random() % grain_level;
		for (int color = -1; color < 3; ++color)
		{
			add_point(level, x, y, color);
			add_point(level, x + 1, y, color);
		}
	}

	return 0;
}

int mmpnm_file::add_jitter(float max_jitter)
{
	if ((max_jitter < 1) || !this || !gray_data || !color_data)
		return -EINVAL;
		
	float x_shift = (random() % (int) (max_jitter * 256)) / 256.0;
	float y_shift = (random() % (int) (max_jitter * 256)) / 256.0;
	fprintf(stderr, "Jitter shifting %f x %f\n", x_shift, y_shift);

	return shift(x_shift, y_shift);
}

int mmpnm_file::add_noise(int max_grain)
{
	if ((max_grain < 1) || !this || !gray_data || !color_data)
		return -EINVAL;

	for (int y = 0, index = 0; y < height; ++y)
		for (int x = 0; x < width; ++x, ++index)
		{
			if (gray_data[index] < 128)
				gray_data[index] += (random() % max_grain);
			else
				gray_data[index] -= (random() % max_grain);

			for (int color = 0; color < 3; ++color)
				if (color_data[(index * 3) + color] < 128)
					color_data[(index * 3) + color] += (random() % max_grain);
				else
					color_data[(index * 3) + color] -= (random() % max_grain);
		}

	return 0;
}

int mmpnm_file::add_pixel(int level, int x, int y, int color)
{
	if (!this || !gray_data || !color_data || (x < 0) || (x >= width) || (y < 0) || (y >= height))
		return -1;

	if (color < 0)
	{
		if ((gray_data[(y * width) + x] + level) > 255)
			gray_data[(y * width) + x] = 255;
		else
			gray_data[(y * width) + x] += level;
	}

	else 
	{
		if ((color_data[(y * width * 3) + x + color] + level) > 255)
			color_data[(y * width * 3) + x + color] = 255;
		else
			color_data[(y * width * 3) + x + color] += level;
	}

	return 0;
}

int mmpnm_file::add_point(int level, float x, float y, int color)
{
	int low_x = (int) floorf(x);
	int low_y = (int) floorf(y);
	float x_interpolate = x - low_x;
	float y_interpolate = y - low_y;

	if (!this || !gray_data || !color_data 
	    || (low_x < 0) || (low_y < 0) 
	    || ((low_x + 1) >= width) || ((low_y + 1) >= height))
		return -1;

	//printf("  Level %d %f %f\n", (unsigned char) (level * y_interpolate * x_interpolate),
	//y_interpolate, x_interpolate);

	add_pixel((int)(level * (1 - x_interpolate) * (1 - y_interpolate)), low_x, low_y, color);
	add_pixel((int)(level * x_interpolate * (1 - y_interpolate)), low_x + 1, low_y, color);
	add_pixel((int)(level * (1 - x_interpolate) * y_interpolate), low_x, low_y + 1, color);
	add_pixel((int)(level * x_interpolate * y_interpolate), low_x + 1, low_y + 1, color);

	return 0;
}

int mmpnm_file::add_to_list(mmpnm_file *&list)
{
	if (!this)
		return -EINVAL;

	if (!list)
	{
		list = this;
		return 0;
	}

	int count = 0;
	mmpnm_file *scan = list;
	while (scan->next)
	{
		scan = scan->next;
		++count;
	}

	scan->next = this;
	return count;	
}

int mmpnm_file::adjust_color(int color, int low_value, int high_value)
{
	if (!this || !color_data)
		return -EINVAL;

	if (low_value < 0)
		low_value = 0;

	if (high_value > 255)
		high_value = 255;

	if (high_value <= low_value)
		return -EINVAL;

	for (int y = 0, index = color; y < height; ++y)
		for (int x = 0; x < width; ++x, index += 3)
		{
			int value = color_data[index];
			value = (value - low_value) * 255 / (high_value - low_value);
			if (value < 0)
				value = 0;
			if (value > 255)
				value = 255;
			color_data[index] = (unsigned char) value;
		}

	return 0;
}

int mmpnm_file::adjust_midpoint(int color, int midpoint)
{
	if (!this || !color_data || !gray_data)
		return -EINVAL;

	if (midpoint <= 0)
		midpoint = 1;

	if (midpoint >= 255)
		midpoint = 254;

	if (color == 3)
	{
		for (int y = 0, index = 0; gray_data && (y < height); ++y)
			for (int x = 0; x < width; ++x, ++index)
			{
				int value = gray_data[index];
				if (value <= midpoint)
					value = 127 * value / midpoint;
				else
					value = 127 + (127 * (value - midpoint) / (255 - midpoint));
	
				if (value < 0)
					value = 0;
				if (value > 255)
					value = 255;
				gray_data[index] = (unsigned char) value;
			}
	}
	else 
	{
		for (int y = 0, index = color; color_data && (y < height); ++y)
			for (int x = 0; x < width; ++x, index += 3)
			{
				int value = color_data[index];
				if (value <= midpoint)
					value = 127 * value / midpoint;
				else
					value = 127 + (127 * (value - midpoint) / (255 - midpoint));

				if (value < 0)
					value = 0;
				if (value > 255)
					value = 255;
				color_data[index] = (unsigned char) value;
			}
	}

	return 0;
}

int mmpnm_file::clear()
{
	if (!this || !gray_data || !color_data)
		return -1;

	memset(gray_data, 0, width * height);
	memset(color_data, 0, width * height * 3);
	return 0;
}

int mmpnm_file::convert_to_frames(mmpnm_file *&frames, mmpnm_framing &parameters)
{
	if (!gray_data || !color_data)
		return -EINVAL;

	// Leave some space at bottom for searching unless using fixed
	// frame height (i.e. ignore sprocket holes)
	
	float bottom = height - (2 * parameters.frame_height);
	if (parameters.ignore_sprocket_holes)
		bottom = height - parameters.frame_height;
		
	// Scan the image from top to bottom
	int frame_count = 0;
	for (float frame_top = parameters.frame_offset; 
			frame_top < bottom; 
			frame_top += parameters.frame_height, ++frame_count)
	{
		char message[256] = "";

		// Find range of levels in frame

		int lo_value = 255;
		int hi_value = 0;
		for (int y = 0; y < (int) parameters.frame_height; ++y)
			for (int x = 0, index = ((int) frame_top + y) * width; x < width; ++x, ++index)
			{
				if (hi_value < gray_data[index])
					hi_value = gray_data[index];

				if (lo_value > gray_data[index])
					lo_value = gray_data[index];
			}


		// Find the left and right edges of the film

		int left_x = 0;
		int right_x = width;
		int threshold = (lo_value + hi_value) / 2;
		for (int y = 0; y < (int) parameters.frame_height; ++y)
		{
			int x = 0;
			int index = ((int) frame_top + y) * width;
			while ((x < width) && (gray_data[index] > threshold))
			{
				++x;
				++index;
			}

			if (x > left_x)
				left_x = x;

			x = width - 1;
			index = (((int) frame_top + y + 1) * width) - 1;
			while ((x > 0) && (gray_data[index] > 128))
			{
				--x;
				--index;
			}

			if (x < right_x)
				right_x = x;
		}


		// Home movies may have sections that are completely overexposed and have 
		// no detectable edges. Assume the frame is in the middle of the image.

		if (right_x < left_x)
		{
			left_x = (width - parameters.film_width) / 2;
			right_x = left_x + parameters.film_width;
			sprintf(&message[strlen(message)], "No detectable edges. ");
		}	

		int minimum_film_width = parameters.film_width - (parameters.film_width / 20);
		
		// Commercial prints may have a clear sprocket area. The detected right 
		// edge can be used for horizontal framing when the clear sprocket area
		// abuts a light image area and there is no consistent left edge.
		
		if (parameters.frame_from_right)
		{
			sprintf(&message[strlen(message)], "Framing from right. ");
			left_x = right_x - parameters.frame_width - parameters.sprocket_area_width;
		}
		
		// Otherwise, the detected left edge should be good. Assume a clear sprocket area
		// if detected width is less expected film width. 

		else if ((right_x - left_x) < minimum_film_width)
		{
			sprintf(&message[strlen(message)], "Clear sprocket area. ");
			left_x -= parameters.sprocket_area_width;
		}
			
		int frame_width = parameters.frame_width - parameters.clip_left - parameters.clip_right;
		int frame_height = (int) parameters.frame_height 
					- parameters.clip_top - parameters.clip_bottom;

		// Find frame by correlation of sprocket holes. When detecting sprocket holes, 
		// even with clear sprocket area, the sprocket hole should be brighter than the 
		// surrounding film area. Scan y values the length of a frame and find the brightest 
		// patch that correlates with an area the size of a sprocket hole.

		// printf("Looking for sprocket at y %d, x %d -> %d, from edge %d, hole %d x %d\n",
		// frame_top, left_x, right_x, parameters.sprocket_to_edge,
		// parameters.sprocket_width, parameters.sprocket_height);

		int correlation = 0;
		int frame_offset = 0;
		if (!parameters.ignore_sprocket_holes)
			for (int offset_scan = 0; 	
				offset_scan < (int) parameters.frame_height; 
				++offset_scan)
			{
				int total = 0;
				for (int y = 0; y < parameters.sprocket_height; ++y)
				{
					int index = (((int) frame_top + offset_scan + y) * width) 
							+ left_x + parameters.sprocket_to_edge;
					for (int x = 0; x < parameters.sprocket_width; ++x, ++index)
						total += gray_data[index];
				}

				if (total > correlation)
				{
					correlation = total;
					frame_offset = offset_scan - (((int) parameters.frame_height 
							- parameters.sprocket_height) / 2);
				}
			}


		// Copy the frame into a new file

		frame_top += frame_offset;
		int frame_x = left_x + parameters.sprocket_area_width + parameters.clip_left;
		int frame_y = (int) frame_top + parameters.clip_top;
 
		mmpnm_file *new_frame = new mmpnm_file;
		new_frame->copy_frame(this, frame_width, frame_height, frame_x, frame_y);

		if (new_frame->gray_data)
			new_frame->add_to_list(frames);
		else // error of some kind, probably range error
			delete new_frame;

		fprintf(stderr, "  Frame at %d/%d (%dx%d - offset %d) %s\n", 
			frame_x, frame_y, frame_width, frame_height, frame_offset, message);
	}

	return frame_count;
}

int mmpnm_file::copy(mmpnm_file *source)
{
	if (!this || !source)
		return -EINVAL;

	memset(this, 0, sizeof(*this));

	width = source->width;
	height = source->height;

	color_data = new unsigned char[width * height * 3];
	memcpy(color_data, source->color_data, width * height * 3);

	gray_data = new unsigned char[width * height];
	memcpy(gray_data, source->gray_data, width * height);

	return width * height;
}

int mmpnm_file::copy_frame(mmpnm_file *source, int frame_width, int frame_height, int left_x, int top_y)
{
	if (!this || !source)
		return -EINVAL;

	if ( (left_x < 0) || (top_y < 0)
		|| ((frame_width + left_x) > source->width) 
		|| ((frame_height + top_y) > source->height))
	{
		fprintf(stderr, "  Problem clipping %dx%d area at %dx%d from %dx%d image\n",
			frame_width, frame_height, left_x, top_y, source->width, source->height);
		return -EINVAL;
	}

	width = frame_width;
	height = frame_height;
	gray_data = new unsigned char[width * height];
	color_data = new unsigned char[width * height * 3];

	int gray_index = 0;
	int color_index = 0;
	for (int y = 0; y < frame_height; ++y)
		for (int x = 0; x < frame_width; ++x, ++gray_index, color_index += 3)
		{
			int source_gray = ((top_y + y) * source->width) + left_x + x;
			int source_color = source_gray * 3;
			gray_data[gray_index] = source->gray_data[source_gray];
			memcpy(&color_data[color_index], &source->color_data[source_color], 3);
		}

	return frame_width * frame_height;
}

int mmpnm_file::crop(int new_width, int new_height)
{
	if (!this || !gray_data || !color_data)
		return -EINVAL;

	if ((width == new_width) && (height == new_height))
		return 0;
		
	unsigned char *old_gray = gray_data;
	unsigned char *old_color = color_data;

	gray_data = (unsigned char*) memset(new unsigned char[new_width * new_height], 0, new_width * new_height);
	color_data = (unsigned char*) memset(new unsigned char[new_width * new_height * 3], 0, new_width * new_height * 3);

	int copy_width = new_width;
	if (copy_width > width)
		copy_width = width;

	int copy_height = new_height;
	if (copy_height > height)
		copy_height = height;

	for (int y = 0; y < copy_height; ++y)
	{
		int source = y * width;
		int destination = y * new_width;
		for (int x = 0; x < copy_width; ++x, ++destination, ++source)
		{
			gray_data[destination] = old_gray[source];
			memcpy(&color_data[destination * 3], &old_color[source * 3], 3);
		}
	}
	
	delete old_color;
	delete old_gray;
	
	width = new_width;
	height = new_height;
	
	return 0;
}

int mmpnm_file::delete_list()
{
	int count = 0;
	mmpnm_file *scan = this;
	
	while (scan)
	{
		mmpnm_file *next_file = scan->next;
		scan->next = NULL;
		delete scan;
		scan = next_file;
		++count;
	}

	return count;
}

int mmpnm_file::find_average_luminance(int color)
{
	if (!this || !color_data || !gray_data || (color < 0) || (color > 3))
		return -EINVAL;

	int total = 0;
	int pixels = width * height;
	if (color < 3)
		pixels *= 3;

	if (color == 3)
		for (int pixel = 0; pixel < pixels; ++pixel)
			total += gray_data[pixel];

	else
		for (int pixel = color; pixel < pixels; pixel += 3)
			total += color_data[pixel];

	return total / pixels;
}

int mmpnm_file::load(char *filename)
{
	int status = 0;
	struct stat file_info;

	if (stat(filename, &file_info) < 0)
	{
		status = -errno;
		perror(filename);
		return status;
	}

	int descriptor = open(filename, O_RDONLY);
	if (descriptor < 0)
	{
		status = -errno;
		perror(filename);
		return status;
	}

	unsigned char *raw_data = new unsigned char [file_info.st_size];
	unsigned char *data_end = &raw_data[file_info.st_size];
	if (read(descriptor, raw_data, file_info.st_size) < 0)
	{
		status = -errno;
		perror(filename);
		close(descriptor);
		return status;
	}

	close(descriptor);

	// Process header text
	int colors = 1;
	int max_value = 255;
	unsigned char *data_start = raw_data;
	for (int argument = 0; (argument < 4) && (data_start < data_end); ++argument)
	{
		if (*data_start == '#')
		{
			--argument; // Comment lines are ignored
			while ((data_start < data_end) && (*data_start >= ' '))
				++data_start;
		}

		else if (argument == 0)
		{
			if (strncmp((char*) data_start, "P6", 2) == 0)
				colors = 3;
			else if (strncmp((char*) data_start, "P5", 2) == 0)
				colors = 1;
			else
			{
				delete raw_data;
				fprintf(stderr, "%s: File is not valid PPM or PGM "
					"(doesn't start with P5 or P6)\n", filename);
				return -EBADF;
			}
		}

		else if (argument == 1)
			width = atoi((char*) data_start);

		else if (argument == 2)
			height = atoi((char*) data_start);

		else if (argument == 3)
			max_value = atoi((char*) data_start);

		// Find end of string
		while ((data_start < data_end) && (*data_start > ' '))
			++data_start;
		
		// Find end of trailing whitespace
		if (argument == 3)
			++data_start;

		else while ((data_start < data_end) && (*data_start <= ' '))
			++data_start;
	}
		
	int pixels = width * height;

	int bytes_per_color = 1;
	if (max_value > 256)
		bytes_per_color = 2;

	int valid_file_length = (data_start - raw_data) + (bytes_per_color * colors * width * height);

	if (valid_file_length > file_info.st_size)
	{
		delete raw_data;
		fprintf(stderr, "File length %d invalid for %d color image, %d bytes/color, %d x %d\n",
			file_info.st_size, colors, bytes_per_color, width, height);
		return -EBADF;
	}
	
	gray_data = new unsigned char[pixels];
	color_data = new unsigned char[pixels * 3];

	int raw_bytes_per_pixel = colors * bytes_per_color;
	int raw_index = 0;
	int gray_index = 0;
	int color_index = 0;

	if (colors == 1) // Grayscale
		for (int pixel = 0; pixel < pixels; ++pixel)
		{
			gray_data[gray_index] = data_start[raw_index];
			color_data[color_index] = gray_data[gray_index];
			color_data[color_index + 1] = gray_data[gray_index];
			color_data[color_index + 2] = gray_data[gray_index];
			raw_index += raw_bytes_per_pixel;
			gray_index += 1;
			color_index += 3;
		}

	else // RGB
		for (int pixel = 0; pixel < pixels; ++pixel)
		{
			gray_data[pixel] = (data_start[raw_index] / 3) 
				+ (data_start[raw_index + bytes_per_color] / 3)
				+ (data_start[raw_index + (2 * bytes_per_color)] / 3);
			color_data[color_index] = data_start[raw_index];
			color_data[color_index + 1] = data_start[raw_index + bytes_per_color];
			color_data[color_index + 2] = data_start[raw_index + (2 * bytes_per_color)];
			raw_index += raw_bytes_per_pixel;
			gray_index += 1;
			color_index += 3;	
		}	

	delete raw_data;

	return pixels;
}

int mmpnm_file::normalize_gray()
{
	if (!this || !gray_data)
		return -EINVAL;

	int pixels = width * height;

	// Find darkest and lightest points
	int low = 255;
	int high = 0;
	for (int pixel = 0; pixel < pixels; ++pixel)
	{
		if (low > gray_data[pixel])
			low = gray_data[pixel];
		if (high < gray_data[pixel])
			high = gray_data[pixel];
	}

	// Normalize to spread over entire spectrum
	double ratio = (high - low) / 256.0;
	if (ratio > 0)
		for (int pixel = 0; pixel < pixels; ++pixel)
		{
			int value = (int)((gray_data[pixel] - low) / ratio);
			if (value > 254)
				value = 254;
			else if (value < 0)
				value = 0;
			gray_data[pixel] = (unsigned char) value;
		}

	return pixels;
}

int mmpnm_file::normalize_color(int color)
{
	if (!this || !color_data || (color < 0) || (color > 2))
		return -EINVAL;

	int pixels = width * height * 3;

	// Find darkest and lightest points
	int low = 255;
	int high = 0;
	for (int pixel = color; pixel < pixels; pixel += 3)
	{
		if (low > color_data[pixel])
			low = color_data[pixel];
		if (high < color_data[pixel])
			high = color_data[pixel];
	}

	// Normalize to spread over entire spectrum
	double ratio = (high - low) / 256.0;
	if (ratio > 0)
		for (int pixel = color; pixel < pixels; pixel += 3)
		{
			int value = (int) ((color_data[pixel] - low) / ratio);
			if (value > 255)
				value = 255;
			if (value < 0)
				value = 0;
			color_data[pixel] = (unsigned char) value;
		}

	return width * height;
}

int mmpnm_file::save(char *filename)
{
	int status = 0;
	char header[256];
	unsigned char *data;
	int data_length = 0;

	if (strstr(filename, ".pgm"))
	{
		sprintf(header, "P5\n%d %d\n255\n", width, height);
		data = gray_data;
		data_length = width * height;
	}
	else
	{
		sprintf(header, "P6\n%d %d\n255\n", width, height);
		data = color_data;
		data_length = width * height * 3;
	}

	int descriptor = open(filename, O_WRONLY | O_TRUNC | O_CREAT, 0644);
	if (descriptor < 0)
	{
		status = -errno;
		perror(filename);
	}

	else if (write(descriptor, header, strlen(header)) < 0)
	{
		status = -errno;
		perror(filename);
	}

	else if (write(descriptor, data, data_length) < 0)
	{
		status = -errno;
		perror(filename);
	}

	if (descriptor >= 0)
		close(descriptor);

	return status;
}

int mmpnm_file::save_frames_to_separate_files(int is_black_and_white, int starting_frame_number)
{
	int frame_number = starting_frame_number;
	for (mmpnm_file *frame = this; frame; ++frame_number, frame = frame->next)
	{
		char filename[256];
		if (is_black_and_white)
			sprintf(filename, "frame%05d.pgm", frame_number);
		else
			sprintf(filename, "frame%05d.ppm", frame_number);

		printf("Saving %s\n", filename);
		int status = frame->save(filename);
		if (status < 0)
			return status;
	}

	return frame_number;
}
		
int mmpnm_file::shift(int x_shift, int y_shift)
{
	unsigned char *old_gray = gray_data;
	unsigned char *old_color = color_data;

	gray_data = (unsigned char*) memcpy(new char[width * height], gray_data, width * height);
	color_data = (unsigned char*) memcpy(new char[width * height * 3], color_data, width * height * 3);

	for (int y = 0; y < height; ++y)
		for (int x = 0; x < width; ++x)
		{
			int shifted_x = (x + width + x_shift) % width;
			int shifted_y = (y + height + y_shift) % height;
			int source = (y * width) + x;
			int destination = (shifted_y * width) + shifted_x;

			gray_data[destination] = old_gray[source];
			memcpy(&color_data[destination * 3], &old_color[source * 3], 3);
		}

	delete old_gray;
	delete old_color;
	
	return 0;
}

int mmpnm_file::shift(float x_shift, float y_shift)
{
	mmpnm_file original;
	original.copy(this);
	clear();
	
	for (int y = 0; y < height; ++y)
		for (int x = 0; x < width; ++x)
		{
			add_point(original.gray_data[(y * width) + x], x + x_shift, y + y_shift, -1);
			add_point(original.color_data[(y * width * 3) + x + 0], x + x_shift, y + y_shift, 0);
			add_point(original.color_data[(y * width * 3) + x + 1], x + x_shift, y + y_shift, 1);
			add_point(original.color_data[(y * width * 3) + x + 2], x + x_shift, y + y_shift, 2);
		}

	return 0;
}
