/***************************************************************************
	mmsuper8.cpp - Converts PNM scans of super 8 film to MOV

			-------------------
        begin           : 5/31/2004
        copyright       : (C) 2004-2005 by Michael Minn
        email           : see http://www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

    Super 8 dimensions (from http://member.melbpc.org.au/~cksm/Formats.html)
    with pixel dimensions at 1200 DPI:

	Film width: 0.314 in (377px)
	Frame height: 0.1667in (200px)
	Frame width: 0.231 (277px)
	Picture width: 0.228 in
	Picture height: 0.163in
	Left margin (sprocket area): 0.058in (70px)
	Right margin (sound stripe): 0.025in (30px)
	Sprocket hole height: 0.045in (54px)
	Sprocket hole width: 0.036in (43px)
	Sprocket to edge: 0.020in (24px)

	www.heroinewarrior.com/quicktime.php3
	smbclient //toshiba1905/frank -I 192.168.1.1 -U frank
	an invalid sample number was found in the movie

***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "mmpnm.h"
#include "mmavi.h"
#include "mmquicktime.h"

/************************************************************************
		Command Line Interface
************************************************************************/

int print_syntax()
{
	fprintf(stderr, "mmsuper8frame [options] scan_file.pnm ...\n\n");
	fprintf(stderr, "	-b\n");
	fprintf(stderr, "		Output will be black and white (i.e. grayscale) (default is color)\n\n");

	fprintf(stderr, "	-resolution dpi\n");
	fprintf(stderr, "		Specifies image resolution for frame sizing based on dimensions of\n");
	fprintf(stderr, "		standard super8 film. Should be specified even if automatic frame \n");
	fprintf(stderr, "		detection is overridden. (default is 1200)\n\n");

	fprintf(stderr, "	-clipleft pixels\n");
	fprintf(stderr, "	-clipright pixels\n");
	fprintf(stderr, "	-cliptop pixels\n");
	fprintf(stderr, "	-clipbottom pixels\n");
	fprintf(stderr, "		Crops a given number of pixels from left/right/top/bottom of each frame\n");
	fprintf(stderr, "		These values can be negative.\n\n");

	fprintf(stderr, "	-nosprocket\n");
	fprintf(stderr, "		Turns off sprocket hole detection and assumes the top of the first\n");
	fprintf(stderr, "		frame is the top of the input image(s) and that subsequent frames\n");
	fprintf(stderr, "		follow at exactly the appropriate frame height for the given resolution.\n");
	fprintf(stderr, "		Useful for scans of commercial prints where sprocket holes are\n");
	fprintf(stderr, "		not detectable in the scan. \n\n");

	fprintf(stderr, "	-frameoffset pixels\n");
	fprintf(stderr, "		When not using sprocket hole detection, the top of the first frame\n");
	fprintf(stderr, "		is assumed to start at the very top of the scan file. This \n");
	fprintf(stderr, "		parameter can be used when the first frame should start at\n");
	fprintf(stderr, "		a lower location. Value is in pixels and can only be positive.\n");
	fprintf(stderr, "		Note that this parameter turns off sprocket hole detection.\n\n");

	fprintf(stderr, "	-frameheight pixels\n");
	fprintf(stderr, "		When not using sprocket hole detection, variation in scanner\n");
	fprintf(stderr, "		accuracy will cause frame drift. This parameter can be tweaked\n");
	fprintf(stderr, "		to adjust the spacing between frames. Values are in pixels.\n");
	fprintf(stderr, "		Default is 200 at 1200 DPI. This parameter turns off sprocket\n");
	fprintf(stderr, "		hole detection: -cliptop and -clipbottom should be used for \n");
	fprintf(stderr, "		adjusting framing and/or modifying the size of the output frame.\n\n");

	fprintf(stderr, "	-framefromright\n");
	fprintf(stderr, "	-framefromleft\n");
	fprintf(stderr, "		Normally a dark left edge is assumed for finding the horizontal\n");
	fprintf(stderr, "		location of a frame. This should also work for commercial\n");
	fprintf(stderr, "		prints that have clear sprocket area since the detected width\n");
	fprintf(stderr, "		will be less than 8mm and the left edge is adjusted accordingly.\n");
	fprintf(stderr, "		However, some commercial prints are framed so the picture area\n");
	fprintf(stderr, "		abuts the clear sprocket area. If the content of the picture \n");
	fprintf(stderr, "		is light, there will be no left edge to detect and horizontal\n");
	fprintf(stderr, "		framing will be incorrect. In those cases, there may be a dark\n");
	fprintf(stderr, "		right border which can be used. -framefromright causes usage\n");
	fprintf(stderr, "		of the right border for framing. The default is framing from\n");
	fprintf(stderr, "		the left and -framefromleft can be used to restore left\n");
	fprintf(stderr, "		framing if -framefromright was given for a previous file in a \n");
	fprintf(stderr, "		list of input files.\n\n");

	fprintf(stderr, "	-sprocket\n");
	fprintf(stderr, "		Turns sprocket hole detection back on if -nosprocket was given\n");
	fprintf(stderr, "		before a previous input file.\n\n");

	fprintf(stderr, "	-frameheight pixels\n");
	fprintf(stderr, "		Specifies distance between each frame (used with -frameoffset to\n");
	fprintf(stderr, "		override automatic frame detection)\n\n");

	fprintf(stderr, "	-frameoffset pixels\n");
	fprintf(stderr, "		Specifies top starting point for frame extraction (used with \n");
	fprintf(stderr, "		-frameheight to override automatic frame detection)\n\n");

	fprintf(stderr, "	-framenumber number\n");
	fprintf(stderr, "		Starting number for output files (default is 00000)\n");

	return 0;
}

int main(int argc, char **argv)
{
	int color = 1;
	int frame_count = 0;
	mmpnm_framing parameters;

	fprintf(stderr, "%s\n", VERSION);

	for (int argument = 1; argument < argc; ++argument)
		if (strncmp(argv[argument], "-h", 2) == 0)
			return print_syntax();

		else if (strcmp(argv[argument], "-b") == 0)
		{
			color = 0;
			fprintf(stderr, "Output will be black and white (grayscale)\n");
		}

		else if (strstr(argv[argument], "-resolution") && ((argument + 1) < argc))
		{
			++argument;
			parameters.set_resolution(atoi(argv[argument]));
		}

		else if (strstr(argv[argument], "-clipleft"))
		{
			++argument;
			parameters.clip_left = ((atoi(argv[argument]) + 1) & 0xfffffffe);
		}

		else if (strstr(argv[argument], "-clipright"))
		{
			++argument;
			parameters.clip_right = ((atoi(argv[argument]) + 1) & 0xfffffffe);
		}

		else if (strstr(argv[argument], "-cliptop"))
		{
			++argument;
			parameters.clip_top = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-clipbottom"))
		{
			++argument;
			parameters.clip_bottom = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-framenumber"))
		{
			++argument;
			frame_count = atoi(argv[argument]);
		}
		
		else if (strstr(argv[argument], "-frameheight"))
		{
			++argument;
			parameters.ignore_sprocket_holes = 1;
			parameters.frame_height = atof(argv[argument]);
			if (parameters.frame_height <= 0)
			{
				fprintf(stderr, "Invalid frame height given %f\n", 
					parameters.frame_height);
				return print_syntax();
			}
		}
		
		else if (strstr(argv[argument], "-frameoffset"))
		{
			++argument;
			parameters.ignore_sprocket_holes = 1;
			parameters.frame_offset = atoi(argv[argument]);
			if (parameters.frame_offset < 0)
			{
				fprintf(stderr, "Invalid frame offset given %d\n", parameters.frame_offset);
				return print_syntax();
			}
		}
		
		else if (strstr(argv[argument], "-framefromright"))
			parameters.frame_from_right = 1;
			
		else if (strstr(argv[argument], "-framefromleft"))
			parameters.frame_from_right = 0;
			
		else if (strstr(argv[argument], "-nosprocket"))
			parameters.ignore_sprocket_holes = 1;

		else if (strstr(argv[argument], "-sprocket"))
			parameters.ignore_sprocket_holes = 0;

		else if (strstr(argv[argument], ".ppm") 
			 || strstr(argv[argument], ".pgm")
			 || strstr(argv[argument], ".pnm"))
		{
			mmpnm_file file;
			if (file.load(argv[argument]) >= 0)
			{
				fprintf(stderr, "%s: %d x %d\n", argv[argument], file.width, file.height);

				if ((file.width < parameters.film_width) || (file.width > file.height))
					fprintf(stderr, "  Input file does not seem to be a strip of frames\n");
				else
				{
					mmpnm_file *frames = NULL;
					file.convert_to_frames(frames, parameters);
					frame_count = frames->save_frames_to_separate_files(!color, frame_count);
					frames->delete_list();
				}
			}
		}
		else
		{
			fprintf(stderr, "Invalid argument: %s\n", argv[argument]);
			return print_syntax();
		}

	fprintf(stderr, "%d frames saved\n", frame_count);

	return 0;
}
